"""
ABSFUYU
-------
A small collection of code

LINKS
-----
- [Home page](https://pypi.org/project/absfuyu/)
- [Documentation](https://absolutewinter.github.io/absfuyu/)

USAGE
-----

Normal import:
>>> import absfuyu
>>> help(absfuyu)

Using in cmd:
``$ fuyu --help``
"""

__title__ = "absfuyu"
__author__ = "AbsoluteWinter"
__license__ = "MIT License"
__version__ = "5.15.1"
__all__ = [
    "core",
    "config",
    "dxt",
    "extra",
    "logger",
    "fun",
    "game",
    "general",
    "pkg_data",
    "sort",
    "tools",
    "util",
    "version",
]
